#!/usr/bin/env python

"""
Data representation functions.

Copyright (C) 2009, 2010, 2011 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from array import array
import operator

# High-level representations.

def convert_sequence(values, op, last_from_preceding):

    """
    Convert the given sequence of 'values' either to a sequence of delta values
    if 'op' is a subtractor of some kind, using the preceding value in the
    sequence as the basis of the calculation of a delta for the current value
    (with 'last_from_preceding' set to a true value), or from a sequence of
    delta values if 'op' is an adder of some kind, using the last result as the
    basis of the calculation of an absolute value (with 'last_from_preceding'
    set to a false value).
    """

    if values:
        new_values = list(values)
        last = new_values[0]
        i = 1
        length = len(new_values)
        while i < length:
            current = new_values[i]
            new_values[i] = op(current, last)

            # Subtracting entries requires the old value to be used.
            # Adding entries requires the new value.

            if last_from_preceding:
                last = current
            else:
                last = new_values[i]

            i += 1

        return new_values
    else:
        return values

# Monotonic operators where an input sequence must always, when combined with an
# operand, produce a result whose elements are all greater than or equal to the
# values of the input sequence (for addition) or all less than or equal to the
# values of the input sequence (for subtraction).
#
# For example: (1, 2) -> (2, 2) -> (2, 3) -> (3, 4)

def op_seq_monotonic(x, y, op):
    return tuple([op(a, b) for a, b in zip(x, y)])

def add_seq_monotonic(x, y):
    return op_seq_monotonic(x, y, operator.add)

def sub_seq_monotonic(x, y):
    return op_seq_monotonic(x, y, operator.sub)

# Operators where an input sequence, when combined with an operand, may produce
# a result whose elements generally obey the properties for a monotonic sequence
# as described above, but where an element in the result may be reset with
# respect to the corresponding element in the input sequence and be less than
# that element (for addition) or greater than that element (for subtraction).
#
# For example: (1, 2) -> (2, 0) -> (2, 3) -> (3, 1)

def add_seq(delta, input):

    "Add 'delta' to 'input'."

    length = min(len(delta), len(input))
    seq = list(delta)[:length]
    i = 0
    while i < length:

        # If the delta is not zero, apply it to the input and retain the rest of
        # the delta sequence as absolute values in the result.

        if delta[i] != 0:
            seq[i] = delta[i] + input[i]
            break

        seq[i] = input[i]
        i += 1
    return tuple(seq)

def sub_seq(current, last):

    "Subtract 'current' from 'last'."

    length = min(len(current), len(last))
    seq = list(current)[:length]
    i = 0
    while i < length:

        # Calculate a suitable delta value. If it is not zero, record it and
        # retain the rest of the current value to reset the sequence in the
        # result.

        replacement = current[i] - last[i]
        if replacement != 0:
            seq[i] = replacement
            break

        seq[i] = 0
        i += 1
    return tuple(seq)

# Data inspection functions.

def is_sequence(value):
    return isinstance(value, (list, tuple))

def sizeof(value):
    return is_sequence(value) and len(value) or 0

# Functions returning operators appropriate for the given structure size.

def get_monotonic_adder(size):
    return size and add_seq_monotonic or operator.add

def get_monotonic_subtractor(size):
    return size and sub_seq_monotonic or operator.sub

def get_adder(size):
    return size and add_seq or operator.add

def get_subtractor(size):
    return size and sub_seq or operator.sub

# Low-level representations.
# Variable-length integer functions.

vint_cache = []
vint_bytes_cache = []

def vint(number):

    "Write 'number' as a variable-length integer."

    try:
        return vint_cache[number]
    except IndexError:
        if number >= 0:
            bytes = array('B')
            _vint_to_array(number, bytes)
            return bytes.tostring()

        # Negative numbers are not supported.

        else:
            raise ValueError, "Number %r is negative." % number

def vint_to_array(number, bytes):

    "Write 'number' as a variable-length integer to 'bytes'."

    try:
        bytes += vint_bytes_cache[number]
    except IndexError:
        if number >= 0:
            _vint_to_array(number, bytes)

        # Negative numbers are not supported.

        else:
            raise ValueError, "Number %r is negative." % number

def _vint_to_array(number, bytes):

    "Write the 'number' to 'bytes' from least to most significant digits."

    while number > 127:
        bytes.append(number & 127 | 128)
        number = number >> 7
    else:
        bytes.append(number)

def vint_from_array(bytes):

    "Read a variable-length integer from 'bytes', returning a number."

    number = 0
    while bytes:
        number <<= 7
        number += bytes.pop() & 127
    return number

def vint_from_array_start(bytes, start):

    """
    Read a variable-length integer from 'bytes', starting at 'start', and
    returning a tuple containing a number and the first position after the
    number.
    """

    length = len(bytes)
    if start == length:
        raise EOFError

    number = 0
    digit = 0
    while start < length:
        x = bytes[start]
        number += (x & 127) << digit
        digit += 7
        start += 1
        if not (x & 128):
            break
    return number, start

# Sequence serialisation.

def sequence_to_array(value, size, bytes):

    "Write the given sequence 'value' with the given 'size' to 'bytes'."

    if size:
        i = 0
        limit = min(len(value), size)
        while i < limit:
            vint_to_array(value[i], bytes)
            i += 1
        while i < size:
            vint_to_array(0, bytes)
    else:
        vint_to_array(value, bytes)

def sequence_from_array(bytes, size, start=0):

    """
    Read a sequence from 'bytes' having the given 'size', returning the sequence
    and the first position after the sequence.
    """

    if size:
        j = 0
        value = []
        while j < size:
            v, start = vint_from_array_start(bytes, start)
            value.append(v)
            j += 1
        return tuple(value), start
    else:
        return vint_from_array_start(bytes, start)

# Variable-length integer cache initialisation.

for i in xrange(0, 65536):
    bytes = array('B')
    _vint_to_array(i, bytes)
    vint_bytes_cache.append(bytes)
    vint_cache.append(bytes.tostring())

# vim: tabstop=4 expandtab shiftwidth=4
